#include <stdbool.h>
#include <string.h>
#include <util/delay.h>     /* for _delay_ms() */
#include "usbdrv.h"
#include "keys.h"

#include <motelib/system.h>
#include <motelib/radio.h>
#include <motelib/led.h>
#include <motelib/button.h>

PROGMEM const char usbHidReportDescriptor[93] =
{
  0x05, 0x01,                    // USAGE_PAGE (Generic Desktop)
  0x09, 0x02,                    // USAGE (Mouse)
  0xa1, 0x01,                    // COLLECTION (Application)
  0x85, 0x01,                    //   REPORT_ID (1)
  0x09, 0x01,                    //   USAGE (Pointer)
  0xA1, 0x00,                    //   COLLECTION (Physical)
  0x05, 0x09,                    //     USAGE_PAGE (Button)
  0x19, 0x01,                    //     USAGE_MINIMUM
  0x29, 0x03,                    //     USAGE_MAXIMUM
  0x15, 0x00,                    //     LOGICAL_MINIMUM (0)
  0x25, 0x01,                    //     LOGICAL_MAXIMUM (1)
  0x95, 0x03,                    //     REPORT_COUNT (3)
  0x75, 0x01,                    //     REPORT_SIZE (1)
  0x81, 0x02,                    //     INPUT (Data,Var,Abs)
  0x95, 0x01,                    //     REPORT_COUNT (1)
  0x75, 0x05,                    //     REPORT_SIZE (5) -- for padding
  0x81, 0x03,                    //     INPUT (Const,Var,Abs)
  0x05, 0x01,                    //     USAGE_PAGE (Generic Desktop)
  0x09, 0x30,                    //     USAGE (X)
  0x09, 0x31,                    //     USAGE (Y)
  0x09, 0x38,                    //     USAGE (Wheel)
  0x15, 0x81,                    //     LOGICAL_MINIMUM (-127)
  0x25, 0x7F,                    //     LOGICAL_MAXIMUM (127)
  0x75, 0x08,                    //     REPORT_SIZE (8)
  0x95, 0x03,                    //     REPORT_COUNT (3)
  0x81, 0x06,                    //     INPUT (Data,Var,Rel)
  0xC0,                          //   END_COLLECTION
  0xC0,                          // END COLLECTION
  0x05, 0x01,                    // USAGE_PAGE (Generic Desktop)
  0x09, 0x06,                    // USAGE (Keyboard)
  0xa1, 0x01,                    // COLLECTION (Application)
  0x85, 0x02,                    //   REPORT_ID (2)
  0x05, 0x07,                    //   USAGE_PAGE (Keyboard)
  0x19, 0xe0,                    //   USAGE_MINIMUM (Keyboard LeftControl)
  0x29, 0xe7,                    //   USAGE_MAXIMUM (Keyboard Right GUI)
  0x15, 0x00,                    //   LOGICAL_MINIMUM (0)
  0x25, 0x01,                    //   LOGICAL_MAXIMUM (1)
  0x75, 0x01,                    //   REPORT_SIZE (1)
  0x95, 0x08,                    //   REPORT_COUNT (8)
  0x81, 0x02,                    //   INPUT (Data,Var,Abs)
  //0x95, 0x01,                    //   REPORT_COUNT (1)
  //0x75, 0x08,                    //   REPORT_SIZE (8)
  //0x81, 0x03,                    //   INPUT (Cnst,Var,Abs)
  0x95, 0x06,                    //   REPORT_COUNT (6)
  0x75, 0x08,                    //   REPORT_SIZE (8)
  0x15, 0x00,                    //   LOGICAL_MINIMUM (0)
  0x25, 0x65,                    //   LOGICAL_MAXIMUM (101)
  0x19, 0x00,                    //   USAGE_MINIMUM (Reserved (no event indicated))
  0x29, 0x65,                    //   USAGE_MAXIMUM (Keyboard Application)
  0x81, 0x00,                    //   INPUT (Data,Ary,Abs)
  0xc0                           // END_COLLECTION
};

#define MOUSE_REPORT_ID     1
#define KEYBOARD_REPORT_ID  2

#define UNKNOW_MODE       0xff
#define MOUSE_MODE        0x01
#define KEYBOARD1_MODE    0x02
#define KEYBOARD2_MODE    0x03

#define LBUTTON_BIT     0x80
#define RBUTTON_BIT     0x40
#define TOP_BIT         0x20
#define RIGHT_BIT       0x10
#define BOTTOM_BIT      0x08
#define LEFT_BIT        0x04

#define MODE_MASK       0x03
#define DATA_MASK       0xfc
#define DIRECTION_MASK  0x3c
#define BUTTON_MASK     0xc0

typedef struct
{
  uint8_t report_id;
  uint8_t buttons;
  int8_t  dx,dy;
  int8_t  dwheel;
} mouse_report_t;

typedef struct
{
  uint8_t report_id;
  uint8_t modifier;
  //uint8_t reserved;
  uint8_t keycode[6];
} keyboard_report_t;

static mouse_report_t     mouseBuffer;
static keyboard_report_t  keyboardBuffer;
static uint8_t            mode;

static uint8_t  idleRate;   /* repeat rate for keyboards, never used for mice */
static bool newRequest;

usbMsgLen_t usbFunctionSetup(uchar data[8])
{
  usbRequest_t *rq = (void *)data;
  
  /* class request type */
  if ((rq->bmRequestType & USBRQ_TYPE_MASK) == USBRQ_TYPE_CLASS)
  {
    if(rq->bRequest == USBRQ_HID_GET_REPORT)
    {
      if (mode == MOUSE_MODE) {
        usbMsgPtr = (void *)&mouseBuffer;
        return sizeof(mouseBuffer);
      } else if (mode == KEYBOARD1_MODE || mode == KEYBOARD2_MODE) {
        usbMsgPtr = (void *)&keyboardBuffer;
        return sizeof(keyboardBuffer);
      }
    }
    else if (rq->bRequest == USBRQ_HID_GET_IDLE)
    {
      usbMsgPtr = &idleRate;
      return 1;
    }
    else if (rq->bRequest == USBRQ_HID_SET_IDLE)
    {
      idleRate = rq->wValue.bytes[1];
    }
  }
  else
  {
    /* no vendor specific requests implemented */
  }
  return 0;   /* default for not implemented requests: return no data back to host */
}

void loop()
{
  
  usbPoll();
  if (usbInterruptIsReady())
  {
    if (newRequest) {
      ledToggle(0);
      if (mode == MOUSE_MODE) {
        usbSetInterrupt((void *)&mouseBuffer, sizeof(mouseBuffer));
        memset(&mouseBuffer, 0, sizeof(mouseBuffer));
        newRequest = false;
      } else if (mode == KEYBOARD1_MODE || mode == KEYBOARD2_MODE) {
        usbSetInterrupt((void *)&keyboardBuffer, sizeof(keyboardBuffer));
        memset(&keyboardBuffer, 0, sizeof(keyboardBuffer));
        newRequest = false;
      }
    } else {
    }
  }
}

static uint8_t oldData;

void recieve(Address src, MessageType type, void *msg, uint8_t len)
{
  if (type == 99) {
    uint8_t data = *(uint8_t*)msg;
    mode = data & MODE_MASK;
    if (mode == MOUSE_MODE) {
      mouseBuffer.report_id = MOUSE_REPORT_ID;
      
      if (data & TOP_BIT)      mouseBuffer.dy -= 1;
      if (data & BOTTOM_BIT)   mouseBuffer.dy += 1;
      if (data & LEFT_BIT)     mouseBuffer.dx -= 1;
      if (data & RIGHT_BIT)    mouseBuffer.dx += 1;
      if (data & LBUTTON_BIT)  mouseBuffer.buttons |= 0x1;
      if (data & RBUTTON_BIT)  mouseBuffer.buttons |= 0x2;

      newRequest = true;
    } else if (mode == KEYBOARD1_MODE && data != oldData) {
      keyboardBuffer.report_id = KEYBOARD_REPORT_ID;
      
      int i = 0;
      if (data & TOP_BIT)      keyboardBuffer.keycode[i++] = KEY_UP;
      if (data & BOTTOM_BIT)   keyboardBuffer.keycode[i++] = KEY_DOWN;
      if (data & LEFT_BIT)     keyboardBuffer.keycode[i++] = KEY_LEFT;
      if (data & RIGHT_BIT)    keyboardBuffer.keycode[i++] = KEY_RIGHT;
      if (data & LBUTTON_BIT)  keyboardBuffer.keycode[i++] = KEY_ENTER;
      if (data & RBUTTON_BIT)  keyboardBuffer.keycode[i]   = KEY_ESC;
      
      newRequest = true;
    } else if (mode == KEYBOARD2_MODE && data != oldData) {
      keyboardBuffer.report_id = KEYBOARD_REPORT_ID;
      
      int i = 0;
      if (data & TOP_BIT)      keyboardBuffer.keycode[i++] = KEY_PAGE_UP;
      if (data & BOTTOM_BIT)   keyboardBuffer.keycode[i++] = KEY_PAGE_DOWN;
      if (data & LEFT_BIT)     keyboardBuffer.keycode[i++] = KEY_HOME;
      if (data & RIGHT_BIT)    keyboardBuffer.keycode[i++] = KEY_END;
      if (data & LBUTTON_BIT)  keyboardBuffer.keycode[i++] = KEY_VOLUME_DOWN;
      if (data & RBUTTON_BIT)  keyboardBuffer.keycode[i]   = KEY_VOLUME_UP;
      
      newRequest = true;
    }
    oldData = data;
  }
}


void boot()
{
  memset(&mouseBuffer, 0, sizeof(mouseBuffer));
  memset(&keyboardBuffer, 0, sizeof(keyboardBuffer));
  mode = UNKNOW_MODE;
  newRequest = false;

  usbInit();
  usbDeviceDisconnect();
  _delay_ms(300);
  usbDeviceConnect();
  
  radioSetRxHandler(recieve);
  setLoopRoutine(loop);
}
